from typing import Type, Dict

from confz.confz_source import ConfZSource, ConfZFileSource, ConfZEnvSource, ConfZCLArgSource, ConfZDataSource
from confz.exceptions import ConfZException
from .cl_arg_loader import CLArgLoader
from .data_loader import DataLoader
from .env_loader import EnvLoader
from .file_loader import FileLoader
from .loader import Loader

_loaders: Dict[Type[ConfZSource], Type[Loader]] = {}


def get_loader(confz_source: Type[ConfZSource]):
    if confz_source in _loaders:
        return _loaders[confz_source]
    else:
        raise ConfZException(f'Unknown config source type "{confz_source}"')


def register_loader(confz_source: Type[ConfZSource], loader: Type[Loader]):
    """Register a :class:`~confz.ConfZSource` with a specific loader. Can be used to extend `ConfZ` with own loaders.

    :param confz_source: The :class:`~confz.ConfZSource` sub-type.
    :param loader: The :class:`~confz.loaders.Loader` sub-type.
    """
    _loaders[confz_source] = loader


register_loader(ConfZFileSource, FileLoader)
register_loader(ConfZEnvSource, EnvLoader)
register_loader(ConfZCLArgSource, CLArgLoader)
register_loader(ConfZDataSource, DataLoader)



from __future__ import annotations  # for sphinx's autodoc_type_aliases

from contextlib import AbstractContextManager
from typing import ClassVar, Type

from pydantic import BaseModel

from .confz_source import ConfZSources
from .exceptions import ConfZException
from .loaders import get_loader


def _load_config(config_kwargs: dict, confz_sources: ConfZSources) -> dict:
    config = config_kwargs.copy()
    if isinstance(confz_sources, list):
        for confz_source in confz_sources:
            loader = get_loader(type(confz_source))
            loader.populate_config(config, confz_source)
    else:
        loader = get_loader(type(confz_sources))
        loader.populate_config(config, confz_sources)
    return config


class ConfZMetaclass(type(BaseModel)):
    """ConfZ Meta Class, inheriting from the pydantic `BaseModel` MetaClass. It would have been cleaner to
    implemented the logic in `__call__` in the ConfZ class itself with `__new__` instead, but pydantic currently
    does not support to overwrite this method."""

    def __call__(cls, config_sources: ConfZSources = None, **kwargs):
        """Called every time an instance of any ConfZ object is created. Injects the config value population and
        singleton mchanism."""
        if config_sources is not None:
            config = _load_config(kwargs, config_sources)
            return super().__call__(**config)

        if cls.CONFIG_SOURCES is not None:
            if len(kwargs) > 0:
                raise ConfZException('Singleton mechanism enabled ("CONFIG_SOURCES" is defined), so keyword arguments '
                                     'are not supported')
            if cls._confz_instance is None:
                config = _load_config(kwargs, cls.CONFIG_SOURCES)
                cls._confz_instance = super().__call__(**config)
            return cls._confz_instance

        return super().__call__(**kwargs)


class SourceChangeManager(AbstractContextManager):
    """Config sources change context manager, allows to change config sources within a controlled context and resets
    everything afterwards."""

    def __init__(self, config_class: Type['ConfZ'], config_sources: ConfZSources):
        self._config_class = config_class
        self._config_sources = config_sources
        self._backup_instance = None
        self._backup_sources = None

    def __enter__(self):
        self._backup_instance = self._config_class._confz_instance
        self._config_class._confz_instance = None

        self._backup_sources = self._config_class.CONFIG_SOURCES
        self._config_class.CONFIG_SOURCES = self._config_sources
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self._config_class._confz_instance = self._backup_instance
        self._config_class.CONFIG_SOURCES = self._backup_sources


class ConfZ(BaseModel, metaclass=ConfZMetaclass):
    """Base class, parent of every config class. Internally wraps :class:`BaseModel` of pydantic and behaves
    transparent except for two cases:

    - If the constructor gets `config_sources` as kwarg, these sources are used as input to enrich the other kwargs.
    - If the class has the class variable `CONFIG_SOURCES` defined, these sources are used as input.

    In the latter case, a singleton mechanism is activated, returning the same config class instance every time the
    constructor is called."""

    CONFIG_SOURCES: ClassVar[ConfZSources] = None   #: Sources to use as input.
    _confz_instance: ClassVar['ConfZ'] = None

    class Config:
        allow_mutation = False

    @classmethod
    def change_config_sources(cls, config_sources: ConfZSources) -> AbstractContextManager:
        """Change the `CONFIG_SOURCES` class variable within a controlled context. Within this context, the sources
        will be different and the singleton reset. This can be useful in unit tests to temporarily
        change a configuration.

        :param config_sources: The temporary config sources for within the context.
        :return: Context manager for change of config sources.
        """
        return SourceChangeManager(cls, config_sources)



import sys

from .loader import Loader
from confz.confz_source import ConfZCLArgSource


class CLArgLoader(Loader):
    """Config loader for command line arguments."""

    @classmethod
    def populate_config(cls, config: dict, confz_source: ConfZCLArgSource):
        cl_args = dict()
        for idx, cl_arg in enumerate(sys.argv[1:]):
            if cl_arg.startswith('--') and idx+2 < len(sys.argv):
                cl_name = cl_arg[2:]
                cl_value = sys.argv[idx+2]

                if confz_source.prefix is not None:
                    if not cl_name.startswith(confz_source.prefix):
                        continue
                    cl_name = cl_name[len(confz_source.prefix):]

                if confz_source.remap is not None and cl_name in confz_source.remap:
                    cl_name = confz_source.remap[cl_name]

                cl_args[cl_name] = cl_value

        cl_args = cls.transform_nested_dicts(cl_args)
        cls.update_dict_recursively(config, cl_args)



from abc import ABC, abstractmethod
from typing import Dict, Any

from confz.confz_source import ConfZSource
from confz.exceptions import ConfZUpdateException


class Loader(ABC):
    """An abstract base class for all config loaders."""

    @classmethod
    def update_dict_recursively(cls, original_dict: Dict, update_dict: Dict):
        """Updates the original dict with the new data. Similar to `dict.update()`, but works with nested dicts.

        :param original_dict: The original dictionary to update in-place.
        :param update_dict: The new data.
        :raises ConfZUpdateException: If dict keys contradict each other.
        """
        for key, value in update_dict.items():
            if isinstance(value, dict) and key in original_dict:
                if not isinstance(original_dict[key], dict):
                    raise ConfZUpdateException(f'Config variables contradict each other: '
                                               f'Key "{key}" is both a value and a nested dict.')
                cls.update_dict_recursively(original_dict[key], value)
            else:
                original_dict[key] = value

    @classmethod
    def transform_nested_dicts(cls, dict_in: Dict[str, Any], separator: str = '.') -> Dict[str, Any]:
        """Transform dictionaries into nested dictionaries, using a separator in the keys as hint.

        :param dict_in: A dictionary with string-keys.
        :param separator: The string used to separate dict keys.
        :return: The transformed dictionary, splitting keys at the separator and creating a new dictionary out of it.
        :raises ConfZUpdateException: If dict keys contradict each other.
        """
        dict_out = dict()
        for key, value in dict_in.items():
            if separator in key:
                inner_keys = key.split(separator)
                dict_inner = dict_out
                for idx, inner_key in enumerate(inner_keys):
                    if idx == len(inner_keys) - 1:
                        dict_inner[inner_key] = value
                    else:
                        if inner_key not in dict_inner:
                            dict_inner[inner_key] = {}
                        else:
                            if not isinstance(dict_inner[inner_key], dict):
                                raise ConfZUpdateException(f'Config variables contradict each other: '
                                                           f'Key "{inner_key}" is both a value and a nested dict.')
                        dict_inner = dict_inner[inner_key]
            else:
                dict_out[key] = value

        return dict_out

    @classmethod
    @abstractmethod
    def populate_config(cls, config: dict, confz_source: ConfZSource):
        """Populate the config-dict with new config arguments based on the source.

        :param config: Config dictionary, gets extended with new arguments
        :param confz_source: Source configuration.
        """



import json
import os
import sys
from pathlib import Path
from typing import Optional

import yaml

from confz.confz_source import ConfZFileSource, FileFormat
from confz.exceptions import ConfZFileException
from .loader import Loader


class FileLoader(Loader):
    """Config loader for config files."""

    @classmethod
    def _get_filename(cls, confz_source: ConfZFileSource) -> Path:
        if confz_source.file is not None:
            file_path = confz_source.file
        elif confz_source.file_from_env is not None:
            if confz_source.file_from_env not in os.environ:
                raise ConfZFileException(f'Environment variable "{confz_source.file_from_env}" is not set.')
            file_path = Path(os.environ[confz_source.file_from_env])
        elif confz_source.file_from_cl is not None:
            if isinstance(confz_source.file_from_cl, int):
                try:
                    file_path = Path(sys.argv[confz_source.file_from_cl])
                except IndexError:
                    raise ConfZFileException(f'Command-line argument number {confz_source.file_from_cl} is not set.')
            else:
                try:
                    idx = sys.argv.index(confz_source.file_from_cl)
                except ValueError:
                    raise ConfZFileException(f'Command-line argument "{confz_source.file_from_cl}" not found.')
                try:
                    file_path = Path(sys.argv[idx+1])
                except IndexError:
                    raise ConfZFileException(f'Command-line argument "{confz_source.file_from_cl}" is not set.')
        else:
            raise ConfZFileException(f'No file source set.')

        if confz_source.folder is not None:
            file_path = confz_source.folder / file_path

        return file_path

    @classmethod
    def _get_format(cls, file_path: Path, file_format: Optional[FileFormat]) -> FileFormat:
        if file_format is not None:
            return file_format

        suffix_formats = {
            '.yml': FileFormat.YAML,
            '.yaml': FileFormat.YAML,
            '.json': FileFormat.JSON
        }
        suffix = file_path.suffix
        try:
            suffix_format = suffix_formats[suffix]
        except KeyError:
            raise ConfZFileException(f'File-ending "{suffix}" is not known. Supported are: '
                                     f'{", ".join(list(suffix_formats.keys()))}.')

        return suffix_format

    @classmethod
    def _read_file(cls, file_path: Path, file_format: FileFormat) -> dict:
        try:
            with file_path.open() as f:
                if file_format == FileFormat.YAML:
                    file_content = yaml.safe_load(f)
                elif file_format == FileFormat.JSON:
                    file_content = json.load(f)
        except OSError as e:
            raise ConfZFileException(f'Could not open config file "{file_path}".') from e

        return file_content

    @classmethod
    def populate_config(cls, config: dict, confz_source: ConfZFileSource):
        file_path = cls._get_filename(confz_source)
        file_format = cls._get_format(file_path, confz_source.format)
        file_content = cls._read_file(file_path, file_format)
        cls.update_dict_recursively(config, file_content)



from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import Optional, Union, List, Dict, Any


@dataclass
class ConfZSource:
    """Source configuration for :class:`~confz.ConfZ` models."""


ConfZSources = Union[ConfZSource, List[ConfZSource]]


class FileFormat(Enum):
    """Enum for file format."""
    JSON = 'json'   #: JSON file format
    YAML = 'yaml'   #: YAML file format


@dataclass
class ConfZFileSource(ConfZSource):
    """Source config for files."""
    file: Optional[Path] = None
    """Specify a config file directly by a path."""
    file_from_env: Optional[str] = None
    """Alternatively, use this environment variable to get the file."""
    file_from_cl: Optional[Union[int, str]] = None
    """Alternatively, use this command line argument to get the file name/path. It can be a specific position 
    (integer, e.g. `1`) or after a specific option (string, e.g. `\\-\\-config-file`). In the latter case, the file
    name must follow after whitespace, an equal sign between argument and value is not supported at the moment."""
    folder: Optional[Path] = None
    """The file specified above can optionally be relative to this folder."""
    format: Optional[FileFormat] = None
    """The format of the config file. If not specified, it will be inferred from the file ending."""


@dataclass
class ConfZEnvSource(ConfZSource):
    """Source config for environment variables. Environment variable names are transformed to
    lowercase and all dashes are replaced by underscores. The definitions below are not case-sensitive and can be
    written with underscore or dash. An exception is `prefix`, which needs to match exactly. Dot-notation
    can be used to access nested configurations."""
    allow_all: bool = False
    """Allow potentially all environment variables to be read as config option."""
    allow: Optional[List[str]] = None
    """Only allow a list of environment variables as input."""
    deny: Optional[List[str]] = None
    """Do not allow to read from environemnt variables in this list. Useful if `allow_all` is set and certain variables 
    should be excluded."""
    prefix: Optional[str] = None
    """The selection above can be narrowed down to a specific prefix, e.g. `CONFIG_`. The variables in the lists above
    or the map below do not need to include this prefix, it is automatically added. This option is especially
    recommended, if ´allow_all´ is set."""
    remap: Optional[Dict[str, str]] = None
    """Certain environment variables can be mapped to config arguments with a different name."""


@dataclass
class ConfZCLArgSource(ConfZSource):
    """Source config for command line arguments. Command line arguments are case-sensitive. Dot-notation can be
    used to access nested configurations. Only command line arguments starting with two dashes (\\-\\-) are considered.
    Between argument and value must be whitespace, an equal sign is not supported at the moment."""
    prefix: Optional[str] = None
    """Optionally, all command line arguments can have a prefix, e.g. `config_`. The prefix does not need to include
    the two dashes at the beginning. The map below does not need to include the prefix, it is automatically added."""
    remap: Optional[Dict[str, str]] = None
    """Certain command line arguments can be mapped to config arguments with a different name. The map does not need to
    include the two dashes at the beginning."""


@dataclass
class ConfZDataSource(ConfZSource):
    """Source config for raw data, i.e. constants. This can be useful for unit-test together with
    :meth:`~confz.ConfZ.change_config_sources` to inject test data into the config."""
    data: Dict[str, Any]
    """All data should go into this (possibly nested) dict."""



from .confz import ConfZ


def _get_sub_classes(cls):
    direct_sub_classes = cls.__subclasses__()

    all_sub_classes = direct_sub_classes.copy()
    for sub_class in direct_sub_classes:
        all_sub_classes.extend(_get_sub_classes(sub_class))

    return all_sub_classes


def validate_all_configs():
    """Instantiates all config classes with a singleton mechanism (`CONFIG_SOURCES` set). This allows to catch
    validation errors early instead of waiting for the first access.

    :raises ConfZException: If any config could not be loaded.
    """
    for config_class in _get_sub_classes(ConfZ):
        if config_class.CONFIG_SOURCES is not None:
            config_class()



from .confz import ConfZ
from .confz_source import ConfZSources, ConfZSource, ConfZFileSource, ConfZEnvSource, ConfZCLArgSource, FileFormat, \
    ConfZDataSource
from .validate import validate_all_configs
